import asyncio
import logging
import sys
from bdb import BdbQuit
from datetime import datetime
from typing import List, Optional

import typer

from ..config.constants import MODEL_SELECTION_CONFIG_PANEL
from ..config.ctx import ElroyContext, get_ctx
from ..config.paths import get_default_config_path, get_default_sqlite_url
from ..io.base import StdIO
from ..io.cli import CliIO
from ..llm.persona import get_persona
from ..logging_config import setup_logging
from ..repository.memories.operations import manually_record_user_memory
from ..repository.user import get_user_id_if_exists
from ..tools.developer import do_print_config
from ..tools.user_preferences import reset_system_persona
from ..tools.user_preferences import set_persona as do_set_persona
from ..utils.utils import datetime_to_string
from .bug_report import create_bug_report_from_exception_if_confirmed
from .chat import (
    handle_message_interactive,
    handle_message_stdio,
    onboard_interactive,
    run_chat,
)
from .options import ElroyOption, get_resolved_params, resolve_model_alias
from .updater import check_latest_version, check_updates

MODEL_ALIASES = ["sonnet", "opus", "gpt4o", "gpt4o_mini", "o1", "o1_mini"]

app = typer.Typer(
    help="Elroy CLI",
    context_settings={
        "obj": {},
    },
    no_args_is_help=False,
)


@app.callback(invoke_without_command=True)
def common(
    typer_ctx: typer.Context,
    config_path: str = typer.Option(
        get_default_config_path(),
        "--config",
        envvar="ELROY_CONFIG_FILE",
        help="YAML config file path. Values override defaults but are overridden by CLI flags and environment variables.",
        rich_help_panel="Basic Configuration",
    ),
    default_assistant_name: str = ElroyOption(
        "default_assistant_name",
        help="Default name for the assistant.",
        rich_help_panel="Basic Configuration",
    ),
    debug: bool = ElroyOption(
        "debug",
        help="Enable fail-fast error handling and verbose logging output.",
        rich_help_panel="Basic Configuration",
    ),
    user_token: str = ElroyOption(
        "user_token",
        help="User token to use for Elroy",
        rich_help_panel="Basic Configuration",
    ),
    custom_tools_path: List[str] = typer.Option(
        [],
        "--custom-tools-path",
        help="Path to custom functions to load",
        show_default=False,
        rich_help_panel="Basic Configuration",
    ),
    # Database Configuration
    database_url: Optional[str] = ElroyOption(
        "database_url",
        default_factory=get_default_sqlite_url,
        help="Valid SQLite or Postgres URL for the database. If Postgres, the pgvector extension must be installed.",
        rich_help_panel="Basic Configuration",
    ),
    # API Configuration
    openai_api_key: Optional[str] = ElroyOption(
        "openai_api_key",
        help="OpenAI API key, required for OpenAI (or OpenAI compatible) models.",
        rich_help_panel=MODEL_SELECTION_CONFIG_PANEL,
    ),
    openai_api_base: Optional[str] = ElroyOption(
        "openai_api_base",
        help="OpenAI API (or OpenAI compatible) base URL.",
        rich_help_panel=MODEL_SELECTION_CONFIG_PANEL,
    ),
    openai_embedding_api_base: Optional[str] = ElroyOption(
        "openai_embedding_api_base",
        help="OpenAI API (or OpenAI compatible) base URL for embeddings.",
        rich_help_panel=MODEL_SELECTION_CONFIG_PANEL,
    ),
    openai_organization: Optional[str] = ElroyOption(
        "openai_organization",
        help="OpenAI (or OpenAI compatible) organization ID.",
        rich_help_panel=MODEL_SELECTION_CONFIG_PANEL,
    ),
    anthropic_api_key: Optional[str] = ElroyOption(
        "anthropic_api_key",
        help="Anthropic API key, required for Anthropic models.",
        rich_help_panel=MODEL_SELECTION_CONFIG_PANEL,
    ),
    # Model Configuration
    chat_model: str = ElroyOption(
        "chat_model",
        help="The model to use for chat completions.",
        rich_help_panel=MODEL_SELECTION_CONFIG_PANEL,
    ),
    inline_tool_calls: bool = ElroyOption(
        "inline_tool_calls",
        help="Whether to enable inline tool calls in the assistant (better for some open source models)",
        rich_help_panel="Basic Configuration",
    ),
    embedding_model: str = ElroyOption(
        "embedding_model",
        help="The model to use for text embeddings.",
        rich_help_panel=MODEL_SELECTION_CONFIG_PANEL,
    ),
    embedding_model_size: int = ElroyOption(
        "embedding_model_size",
        help="The size of the embedding model.",
        rich_help_panel=MODEL_SELECTION_CONFIG_PANEL,
    ),
    enable_caching: bool = ElroyOption(
        "enable_caching",
        help="Whether to enable caching for the LLM, both for embeddings and completions.",
        rich_help_panel=MODEL_SELECTION_CONFIG_PANEL,
    ),
    # Context Management
    max_assistant_loops: int = ElroyOption(
        "max_assistant_loops",
        help="Maximum number of loops the assistant can run before tools are temporarily made unvailable (returning for the next user message).",
        rich_help_panel="Context Management",
    ),
    context_refresh_trigger_tokens: int = ElroyOption(
        "context_refresh_trigger_tokens",
        help="Number of tokens that triggers a context refresh and compresion of messages in the context window.",
        rich_help_panel="Context Management",
    ),
    context_refresh_target_tokens: int = ElroyOption(
        "context_refresh_target_tokens",
        help="Target number of tokens after context refresh / context compression, how many tokens to aim to keep in context.",
        rich_help_panel="Context Management",
    ),
    max_context_age_minutes: float = ElroyOption(
        "max_context_age_minutes",
        help="Maximum age in minutes to keep. Messages older tha this will be dropped from context, regardless of token limits",
        rich_help_panel="Context Management",
    ),
    min_convo_age_for_greeting_minutes: Optional[float] = ElroyOption(
        "min_convo_age_for_greeting_minutes",
        help="Minimum age in minutes of conversation before the assistant will offer a greeting on login. 0 means assistant will offer greeting each time. To disable greeting, set enable_assistant_greeting=False (This will override any value for min_convo_age_for_greeting_minutes)",
        rich_help_panel="Context Management",
    ),
    enable_assistant_greeting: bool = ElroyOption(
        "enable_assistant_greeting",
        help="Whether to allow the assistant to send the first message",
        rich_help_panel="Context Management",
    ),
    # Memory Consolidation
    memories_between_consolidation: int = ElroyOption(
        "memories_between_consolidation",
        help="How many memories to create before triggering a memory consolidation operation.",
        rich_help_panel="Memory Consolidation",
    ),
    l2_memory_relevance_distance_threshold: float = ElroyOption(
        "l2_memory_relevance_distance_threshold",
        help="L2 distance threshold for memory relevance.",
        rich_help_panel="Memory Consolidation",
    ),
    memory_cluster_similarity_threshold: float = ElroyOption(
        "memory_cluster_similarity_threshold",
        help="Threshold for memory cluster similarity.",
        rich_help_panel="Memory Consolidation",
    ),
    max_memory_cluster_size: int = ElroyOption(
        "max_memory_cluster_size",
        help="The maximum number of memories that can be consolidated into a single memory at once.",
        rich_help_panel="Memory Consolidation",
    ),
    min_memory_cluster_size: int = ElroyOption(
        "min_memory_cluster_size",
        help="The minimum number of memories that can be consolidated into a single memory at once.",
        rich_help_panel="Memory Consolidation",
    ),
    initial_context_refresh_wait_seconds: int = ElroyOption(
        "initial_context_refresh_wait_seconds",
        help="Initial wait time in seconds after login before the initial context refresh and compression.",
        rich_help_panel="Memory Consolidation",
    ),
    # UI Configuration
    show_internal_thought: bool = ElroyOption(
        "show_internal_thought",
        help="Show the assistant's internal thought monologue like memory consolidation and internal reflection.",
        rich_help_panel="UI Configuration",
    ),
    system_message_color: str = ElroyOption(
        "system_message_color",
        help="Color for system messages.",
        rich_help_panel="UI Configuration",
    ),
    user_input_color: str = ElroyOption(
        "user_input_color",
        help="Color for user input.",
        rich_help_panel="UI Configuration",
    ),
    assistant_color: str = ElroyOption(
        "assistant_color",
        help="Color for assistant output.",
        rich_help_panel="UI Configuration",
    ),
    warning_color: str = ElroyOption(
        "warning_color",
        help="Color for warning messages.",
        rich_help_panel="UI Configuration",
    ),
    internal_thought_color: str = ElroyOption(
        "internal_thought_color",
        help="Color for internal thought messages.",
        rich_help_panel="UI Configuration",
    ),
    sonnet: bool = typer.Option(
        False,
        "--sonnet",
        help="Use Anthropic's Sonnet model",
        show_default=False,
        rich_help_panel=MODEL_SELECTION_CONFIG_PANEL,
    ),
    opus: bool = typer.Option(
        False,
        "--opus",
        help="Use Anthropic's Opus model",
        show_default=False,
        rich_help_panel=MODEL_SELECTION_CONFIG_PANEL,
    ),
    gpt4o: bool = typer.Option(
        False,
        "--4o",
        help="Use OpenAI's GPT-4o model",
        show_default=False,
        rich_help_panel=MODEL_SELECTION_CONFIG_PANEL,
    ),
    gpt4o_mini: bool = typer.Option(
        False,
        "--4o-mini",
        help="Use OpenAI's GPT-4o-mini model",
        show_default=False,
        rich_help_panel=MODEL_SELECTION_CONFIG_PANEL,
    ),
    o1: bool = typer.Option(
        False,
        "--o1",
        help="Use OpenAI's o1 model",
        show_default=False,
        rich_help_panel=MODEL_SELECTION_CONFIG_PANEL,
    ),
    o1_mini: bool = typer.Option(
        False,
        "--o1-mini",
        help="Use OpenAI's o1-mini model",
        show_default=False,
        rich_help_panel=MODEL_SELECTION_CONFIG_PANEL,
    ),
):
    """Common parameters."""

    for m in MODEL_ALIASES:
        if typer_ctx.params.get(m):
            logging.info(f"Model alias {m} selected")
            resolved = resolve_model_alias(m)
            if not resolved:
                logging.warning("Model alias not found")
            else:
                typer_ctx.params["chat_model"] = resolved
        del typer_ctx.params[m]

    typer_ctx.obj = ElroyContext(
        **get_resolved_params(**typer_ctx.params),
    )

    setup_logging()

    if typer_ctx.invoked_subcommand is None:
        chat(typer_ctx)


@app.command(name="chat")
def chat(typer_ctx: typer.Context):
    """Opens an interactive chat session."""
    ctx = get_ctx(typer_ctx)
    with ctx.dbsession():
        if sys.stdin.isatty():
            check_updates()
            try:
                if not get_user_id_if_exists(ctx.db, ctx.user_token):
                    asyncio.run(onboard_interactive(ctx))
                asyncio.run(run_chat(ctx))
            except BdbQuit:
                logging.info("Exiting...")
            except EOFError:
                logging.info("Exiting...")
            except Exception as e:
                create_bug_report_from_exception_if_confirmed(ctx, e)
        else:
            message = sys.stdin.read()
            handle_message_stdio(ctx, StdIO(), message, None)


@app.command(name="message")
def message(
    typer_ctx: typer.Context,
    message: Optional[str] = typer.Argument(..., help="The message to process"),
    tool: str = typer.Option(
        None,
        "--tool",
        "-t",
        help="Specifies the tool to use in responding to a message",
    ),
):
    """Process a single message and exit."""
    ctx = get_ctx(typer_ctx)
    with ctx.dbsession():
        if sys.stdin.isatty() and not message:
            io = ctx.io
            assert isinstance(io, CliIO)
            handle_message_interactive(ctx, io, tool)
        else:
            assert message
            handle_message_stdio(ctx, StdIO(), message, tool)


@app.command(name="print-tools")
def print_tools(
    typer_ctx: typer.Context,
    tool: Optional[str] = typer.Argument(None, help="Tool to print schema for"),
):
    """Prints the schema for a tool and exits."""
    ctx = get_ctx(typer_ctx)
    ctx.io.print(ctx.tool_registry.get_schemas())  # type: ignore


@app.command(name="remember")
def remember(
    typer_ctx: typer.Context,
    text: Optional[str] = typer.Argument(
        None,
        help="Text to remember. If not provided, will prompt interactively",
    ),
):
    """Create a new memory from text or interactively."""
    ctx = get_ctx(typer_ctx)
    with ctx.dbsession():
        if not get_user_id_if_exists(ctx.db, ctx.user_token):
            ctx.io.warning("Creating memory for new user")

        if text:
            memory_name = f"Memory from CLI, created {datetime_to_string(datetime.now())}"
            manually_record_user_memory(ctx, text, memory_name)
            ctx.io.info(f"Memory created: {memory_name}")
            raise typer.Exit()

        elif sys.stdin.isatty():
            io = ctx.io
            assert isinstance(io, CliIO)
            memory_text = asyncio.run(io.prompt_user("Enter the memory text:"))
            memory_text += f"\nManually entered memory, at: {datetime_to_string(datetime.now())}"
            # Optionally get memory name
            memory_name = asyncio.run(io.prompt_user("Enter memory name (optional, press enter to skip):"))
            try:
                manually_record_user_memory(ctx, memory_text, memory_name)
                ctx.io.info(f"Memory created: {memory_name}")
                raise typer.Exit()
            except ValueError as e:
                ctx.io.warning(f"Error creating memory: {e}")
                raise typer.Exit(1)
        else:
            memory_text = sys.stdin.read()
            metadata = "Memory ingested from stdin\n" f"Ingested at: {datetime_to_string(datetime.now())}\n"
            memory_text = f"{metadata}\n{memory_text}"
            memory_name = f"Memory from stdin, ingested {datetime_to_string(datetime.now())}"
            manually_record_user_memory(ctx, memory_text, memory_name)
            ctx.io.info(f"Memory created: {memory_name}")
            raise typer.Exit()


@app.command(name="list-models")
def list_models():
    """Lists supported chat models and exits."""
    from ..config.models import (
        get_supported_anthropic_models,
        get_supported_openai_models,
    )

    for m in get_supported_openai_models():
        print(f"{m} (OpenAI)")
    for m in get_supported_anthropic_models():
        print(f"{m} (Anthropic)")
    raise typer.Exit()


@app.command(name="print-config")
def print_config(
    typer_ctx: typer.Context,
    show_secrets: bool = typer.Option(
        False,
        "--show-secrets",
        help="Whether to show secret values in output",
    ),
):
    """Shows current configuration and exits."""
    ctx = get_ctx(typer_ctx)
    ctx.io.print(do_print_config(ctx, show_secrets))


@app.command()
def version():
    """Show version and exit."""
    current_version, latest_version = check_latest_version()
    if latest_version > current_version:
        typer.echo(f"Elroy version: {current_version} (newer version {latest_version} available)")
        typer.echo("\nTo upgrade, run:")
        typer.echo(f"    pip install --upgrade elroy=={latest_version}")
    else:
        typer.echo(f"Elroy version: {current_version} (up to date)")

    raise typer.Exit()


@app.command(name="set-persona")
def set_persona(
    typer_ctx: typer.Context,
    persona: str = typer.Argument(..., help="Persona text to set"),
):
    """Set a custom persona for the assistant."""
    ctx = get_ctx(typer_ctx)
    with ctx.dbsession():
        if get_user_id_if_exists(ctx.db, ctx.user_token):
            logging.info(f"No user found for token {ctx.user_token}, creating one")
        do_set_persona(ctx, persona)
        raise typer.Exit()


@app.command(name="reset-persona")
def reset_persona(typer_ctx: typer.Context):
    """Removes any custom persona, reverting to the default."""
    ctx = get_ctx(typer_ctx)
    with ctx.dbsession():
        if not get_user_id_if_exists(ctx.db, ctx.user_token):
            logging.warning(f"No user found for token {ctx.user_token}, so no persona to clear")
            return typer.Exit()
        else:
            reset_system_persona(ctx)
        raise typer.Exit()


@app.command(name="show-persona")
def show_persona(typer_ctx: typer.Context):
    """Print the system persona and exit."""
    ctx = get_ctx(typer_ctx)
    with ctx.dbsession():
        print(get_persona(ctx))
        raise typer.Exit()


if __name__ == "__main__":
    app()



import logging
import traceback
from functools import partial
from typing import Iterator, List, Optional, Union

from toolz import juxt, merge, pipe
from toolz.curried import do, filter, map, remove, tail

from ..config.constants import (
    ASSISTANT,
    SYSTEM,
    TOOL,
    USER,
    MissingAssistantToolCallError,
    MissingToolCallMessageError,
    RecoverableToolError,
)
from ..config.ctx import ElroyContext
from ..db.db_models import FunctionCall
from ..llm.client import generate_chat_completion_message, get_embedding
from ..llm.stream_parser import (
    AssistantInternalThought,
    AssistantResponse,
    AssistantToolResult,
)
from ..repository.data_models import ContextMessage
from ..repository.embeddings import get_most_relevant_goal, get_most_relevant_memory
from ..repository.message import (
    MemoryMetadata,
    add_context_messages,
    get_context_messages,
    is_system_instruction,
    replace_context_messages,
)
from ..tools.function_caller import ERROR_PREFIX
from ..utils.utils import last_or_none, logged_exec_time
from .context import get_refreshed_system_message


def process_message(
    role: str, ctx: ElroyContext, msg: str, force_tool: Optional[str] = None
) -> Iterator[Union[AssistantResponse, AssistantInternalThought, AssistantToolResult]]:
    assert role in [USER, ASSISTANT, SYSTEM]

    context_messages = pipe(
        get_context_messages(ctx),
        partial(validate, ctx),
        list,
    )

    new_msgs = [ContextMessage(role=role, content=msg, chat_model=None)]
    new_msgs += get_relevant_memories(ctx, context_messages + new_msgs)

    loops = 0
    while True:
        function_calls: List[FunctionCall] = []
        tool_context_messages: List[ContextMessage] = []

        stream = generate_chat_completion_message(
            chat_model=ctx.chat_model,
            context_messages=context_messages + new_msgs,
            tool_schemas=ctx.tool_registry.get_schemas(),
            enable_tools=(not ctx.chat_model.inline_tool_calls) and loops <= ctx.max_assistant_loops,
            force_tool=force_tool,
        )
        for stream_chunk in stream.process():
            if isinstance(stream_chunk, (AssistantResponse, AssistantInternalThought)):
                yield stream_chunk
            elif isinstance(stream_chunk, FunctionCall):
                pipe(
                    stream_chunk,
                    do(function_calls.append),
                    lambda x: ContextMessage(
                        role=TOOL,
                        tool_call_id=x.id,
                        content=exec_function_call(ctx, x),
                        chat_model=ctx.chat_model.name,
                    ),
                    tool_context_messages.append,
                )
        new_msgs.append(
            ContextMessage(
                role=ASSISTANT,
                content=stream.get_full_text(),
                tool_calls=(None if not function_calls else [f.to_tool_call() for f in function_calls]),
                chat_model=ctx.chat_model.name,
            )
        )

        if force_tool:
            assert len(tool_context_messages) >= 1
            if len(tool_context_messages) > 1:
                logging.warning(f"With force tool {force_tool}, expected one tool message, but found {len(tool_context_messages)}")

            new_msgs += tool_context_messages
            add_context_messages(ctx, new_msgs)

            content = tool_context_messages[-1].content
            assert isinstance(content, str)
            yield AssistantToolResult(content)
            break

        elif tool_context_messages:
            new_msgs += tool_context_messages
        else:
            add_context_messages(ctx, new_msgs)
            break
        loops += 1


def exec_function_call(ctx: ElroyContext, function_call: FunctionCall) -> str:
    ctx.io.print(function_call)
    function_to_call = ctx.tool_registry.get(function_call.function_name)
    if not function_to_call:
        return f"Function {function_call.function_name} not found"

    try:
        result = pipe(
            {"ctx": ctx} if "ctx" in function_to_call.__code__.co_varnames else {},
            lambda d: merge(function_call.arguments, d),
            lambda args: function_to_call.__call__(**args),
            lambda result: str(result) if result is not None else "Success",
        )

    except RecoverableToolError as e:
        result = f"Tool error: {e}"

    except Exception as e:
        return pipe(
            f"Failed function call:\n{function_call}\n\n" + "".join(traceback.format_exception(type(e), e, e.__traceback__)),
            do(ctx.io.warning),
            ERROR_PREFIX.__add__,
        )
    assert isinstance(result, str)
    ctx.io.info(result)
    return result


def validate(ctx: ElroyContext, context_messages: List[ContextMessage]) -> List[ContextMessage]:
    messages = pipe(
        context_messages,
        partial(_validate_system_instruction_correctly_placed, ctx),
        partial(_validate_assistant_tool_calls_followed_by_tool, ctx.debug),
        partial(_validate_tool_messages_have_assistant_tool_call, ctx.debug),
        lambda msgs: (msgs if not ctx.chat_model.ensure_alternating_roles else validate_first_user_precedes_first_assistant(msgs)),
        list,
    )

    if messages != context_messages:
        logging.info("Context messages have been repaired")
        replace_context_messages(ctx, messages)
    return messages


def validate_first_user_precedes_first_assistant(context_messages: List[ContextMessage]) -> List[ContextMessage]:
    user_and_assistant_messages = [m for m in context_messages if m.role in [USER, ASSISTANT]]

    if user_and_assistant_messages and user_and_assistant_messages[0].role != USER:
        logging.info("First non-system message is not user message, repairing by inserting user message")

        context_messages = [
            context_messages[0],
            ContextMessage(role=USER, content="The user has begun the converstaion", chat_model=None),
        ] + context_messages[1:]
    return context_messages


def _validate_system_instruction_correctly_placed(ctx: ElroyContext, context_messages: List[ContextMessage]) -> List[ContextMessage]:
    validated_messages = []
    for idx, message in enumerate(context_messages):
        if idx == 0 and not is_system_instruction(message):
            logging.info(f"First message is not system instruction, repairing by inserting system instruction")
            validated_messages += [
                get_refreshed_system_message(ctx, context_messages),
                message,
            ]
        elif idx != 0 and is_system_instruction(message):
            logging.error("Found system message in non-first position, repairing by dropping message")
            continue
        else:
            validated_messages.append(message)
    return validated_messages


def _validate_assistant_tool_calls_followed_by_tool(debug_mode: bool, context_messages: List[ContextMessage]) -> List[ContextMessage]:
    """
    Validates that any assistant message with non-empty tool_calls is followed by corresponding tool messages.
    """

    for idx, message in enumerate(context_messages):
        if (message.role == ASSISTANT and message.tool_calls is not None) and (
            idx == len(context_messages) - 1 or context_messages[idx + 1].role != TOOL
        ):
            if debug_mode:
                raise MissingToolCallMessageError()
            else:
                logging.error(
                    f"Assistant message with tool_calls not followed by tool message: ID = {message.id}, repairing by removing tool_calls"
                )
                message.tool_calls = None
    return context_messages


def _validate_tool_messages_have_assistant_tool_call(debug_mode: bool, context_messages: List[ContextMessage]) -> List[ContextMessage]:
    """
    Validates that all tool messages have a preceding assistant message with the corresponding tool_calls.
    """

    validated_context_messages = []
    for idx, message in enumerate(context_messages):
        if message.role == TOOL and not _has_assistant_tool_call(message.tool_call_id, context_messages[:idx]):
            if debug_mode:
                raise MissingAssistantToolCallError(f"Message id: {message.id}")
            else:
                logging.warning(
                    f"Tool message without preceding assistant message with tool_calls: ID = {message.id}. Repairing by removing tool message"
                )
                continue
        else:
            validated_context_messages.append(message)

    return validated_context_messages


def _has_assistant_tool_call(tool_call_id: Optional[str], context_messages: List[ContextMessage]) -> bool:
    """
    Assistant tool call message must be in the most recent assistant message
    """
    if not tool_call_id:
        logging.warning("Tool call ID is None")
        return False

    return pipe(
        context_messages,
        filter(lambda x: x.role == ASSISTANT),
        last_or_none,
        lambda msg: msg.tool_calls or [] if msg else [],
        map(lambda x: x.id),
        filter(lambda x: x == tool_call_id),
        any,
    )


@logged_exec_time
def get_relevant_memories(ctx: ElroyContext, context_messages: List[ContextMessage]) -> List[ContextMessage]:
    from ..repository.embeddable import is_in_context

    message_content = pipe(
        context_messages,
        remove(lambda x: x.role == SYSTEM),
        tail(4),
        map(lambda x: f"{x.role}: {x.content}" if x.content else None),
        remove(lambda x: x is None),
        list,
        "\n".join,
    )

    if not message_content:
        return []

    assert isinstance(message_content, str)

    new_memory_messages = pipe(
        message_content,
        partial(get_embedding, ctx.embedding_model),
        lambda x: juxt(get_most_relevant_goal, get_most_relevant_memory)(ctx, x),
        filter(lambda x: x is not None),
        remove(partial(is_in_context, context_messages)),
        map(
            lambda x: ContextMessage(
                role=SYSTEM,
                memory_metadata=[MemoryMetadata(memory_type=x.__class__.__name__, id=x.id, name=x.get_name())],
                content="Information recalled from assistant memory: " + x.to_fact(),
                chat_model=None,
            )
        ),
        list,
    )

    return new_memory_messages



import logging
from functools import partial
from typing import List, Type

from sqlmodel import select
from toolz import pipe
from toolz.curried import filter

from ..config.constants import SYSTEM
from ..config.ctx import ElroyContext
from ..db.db_models import EmbeddableSqlModel
from .data_models import ContextMessage
from .message import (
    MemoryMetadata,
    add_context_messages,
    get_context_messages,
    remove_context_messages,
)


def is_in_context_message(memory: EmbeddableSqlModel, context_message: ContextMessage) -> bool:
    if not context_message.memory_metadata:
        return False
    return any(x.memory_type == memory.__class__.__name__ and x.id == memory.id for x in context_message.memory_metadata)


def remove_from_context(ctx: ElroyContext, memory: EmbeddableSqlModel):
    pipe(
        get_context_messages(ctx),
        filter(partial(is_in_context_message, memory)),
        list,
        partial(remove_context_messages, ctx),
    )


def add_to_context(ctx: ElroyContext, memory: EmbeddableSqlModel) -> None:
    memory_id = memory.id
    assert memory_id

    context_messages = get_context_messages(ctx)

    if is_in_context(context_messages, memory):
        logging.info(f"Memory of type {memory.__class__.__name__} with id {memory_id} already in context.")
    else:
        add_context_messages(
            ctx,
            [
                ContextMessage(
                    role=SYSTEM,
                    memory_metadata=[MemoryMetadata(memory_type=memory.__class__.__name__, id=memory_id, name=memory.get_name())],
                    content=memory.to_fact(),
                    chat_model=None,
                )
            ],
        )


def is_in_context(context_messages: List[ContextMessage], memory: EmbeddableSqlModel) -> bool:
    return any(is_in_context_message(memory, x) for x in context_messages)


def add_to_current_context_by_name(ctx: ElroyContext, name: str, memory_type: Type[EmbeddableSqlModel]) -> str:
    item = ctx.db.exec(select(memory_type).where(memory_type.name == name)).first()  # type: ignore

    if item:
        add_to_context(ctx, item)
        return f"{memory_type.__name__} '{name}' added to context."
    else:
        return f"{memory_type.__name__} '{name}' not found."


def drop_from_context_by_name(ctx: ElroyContext, name: str, memory_type: Type[EmbeddableSqlModel]) -> str:
    item = ctx.db.exec(select(memory_type).where(memory_type.name == name)).first()  # type: ignore

    if item:
        remove_from_context(ctx, item)
        return f"{memory_type.__name__} '{name}' dropped from context."
    else:
        return f"{memory_type.__name__} '{name}' not found."



import json
from dataclasses import dataclass
from typing import Dict, Iterator, List, Optional

from ..config.config import ChatModel
from ..db.db_models import FunctionCall


@dataclass
class PartialToolCall:
    id: str
    model: str
    function_name: str = ""
    arguments: str = ""
    type: str = "function"
    is_complete: bool = False

    from litellm.types.utils import ChatCompletionDeltaToolCall

    def update(self, delta: ChatCompletionDeltaToolCall) -> Optional[FunctionCall]:
        from litellm.types.utils import ChatCompletionDeltaToolCall

        if self.is_complete:
            raise ValueError("PartialToolCall is already complete")

        assert isinstance(delta, ChatCompletionDeltaToolCall), f"Expected ChoiceDeltaToolCall, got {type(delta)}"
        assert delta.function
        if delta.function.name:
            self.function_name += delta.function.name
        if delta.function.arguments:
            self.arguments += delta.function.arguments

        # Check if we have a complete JSON object for arguments
        try:
            function_call = FunctionCall(
                id=self.id,
                function_name=self.function_name,
                arguments=json.loads(self.arguments),
            )
            self.is_complete = True
            return function_call
        except json.JSONDecodeError:
            return None


class OpenAIToolCallAccumulator:
    from litellm.types.utils import ChatCompletionDeltaToolCall

    def __init__(self, chat_model: ChatModel):
        self.chat_model = chat_model
        self.tool_calls: Dict[int, PartialToolCall] = {}
        self.last_updated_index: Optional[int] = None

    def update(self, delta_tool_calls: Optional[List[ChatCompletionDeltaToolCall]]) -> Iterator[FunctionCall]:
        for delta in delta_tool_calls or []:
            if delta.index not in self.tool_calls:
                if (
                    self.last_updated_index is not None
                    and self.last_updated_index in self.tool_calls
                    and self.last_updated_index != delta.index
                ):
                    raise ValueError("New tool call started, but old one is not yet complete")
                assert delta.id
                self.tool_calls[delta.index] = PartialToolCall(id=delta.id, model=self.chat_model.name)

            completed_tool_call = self.tool_calls[delta.index].update(delta)
            if completed_tool_call:
                self.tool_calls.pop(delta.index)
                yield completed_tool_call
            else:
                self.last_updated_index = delta.index



import asyncio
from datetime import datetime

from pytz import UTC
from sqlmodel import select

from ..config.constants import USER
from ..config.ctx import ElroyContext
from ..db.db_models import Message
from ..messaging.context import context_refresh, is_context_refresh_needed
from ..tools.user_preferences import get_user_preferred_name
from ..utils.utils import datetime_to_string


def refresh_context_if_needed(ctx: ElroyContext):
    if is_context_refresh_needed(ctx):
        asyncio.run(context_refresh(ctx))


def get_user_logged_in_message(ctx: ElroyContext) -> str:
    preferred_name = get_user_preferred_name(ctx)

    if preferred_name == "Unknown":
        preferred_name = "User (preferred name unknown)"

    local_tz = datetime.now().astimezone().tzinfo

    # Get start of today in local timezone
    today_start = datetime.now(local_tz).replace(hour=0, minute=0, second=0, microsecond=0)

    # Convert to UTC for database comparison
    today_start_utc = today_start.astimezone(UTC)

    earliest_today_msg = ctx.db.exec(
        select(Message)
        .where(Message.user_id == ctx.user_id)
        .where(Message.role == USER)
        .where(Message.created_at >= today_start_utc)
        .order_by(Message.created_at)  # type: ignore
        .limit(1)
    ).first()

    if earliest_today_msg:
        # Convert UTC time to local timezone for display
        local_time = earliest_today_msg.created_at.replace(tzinfo=UTC).astimezone(local_tz)
        today_summary = f"I first started chatting with {preferred_name} today at {local_time.strftime('%I:%M %p')}."
    else:
        today_summary = f"I haven't chatted with {preferred_name} yet today. I should offer a brief greeting."

    return f"{preferred_name} has logged in. The current time is {datetime_to_string(datetime.now().astimezone())}. {today_summary}"



from contextlib import contextmanager
from datetime import timedelta
from functools import cached_property
from pathlib import Path
from types import SimpleNamespace
from typing import Any, Callable, Generator, List, Optional, TypeVar

import typer

from ..db.db_manager import DbManager
from ..db.postgres.postgres_manager import PostgresManager
from ..db.sqlite.sqlite_manager import SqliteManager
from ..io.cli import CliIO
from ..repository.user import create_user_id, get_user_id_if_exists
from .config import ChatModel, EmbeddingModel, get_chat_model, get_embedding_model
from .paths import get_default_config_path


class ElroyContext:
    from ..io.base import ElroyIO
    from ..io.cli import CliIO

    _db: Optional[DbManager] = None
    _io: Optional[ElroyIO] = None

    def __init__(
        self,
        *,
        # Basic Configuration
        config_path: Optional[str] = None,
        database_url: str,
        show_internal_thought: bool,
        system_message_color: str,
        assistant_color: str,
        user_input_color: str,
        warning_color: str,
        internal_thought_color: str,
        user_token: str,
        custom_tools_path: List[str] = [],
        # API Configuration
        openai_api_key: Optional[str] = None,
        openai_api_base: Optional[str] = None,
        openai_embedding_api_base: Optional[str] = None,
        openai_organization: Optional[str] = None,
        anthropic_api_key: Optional[str] = None,
        # Model Configuration
        chat_model: str,
        embedding_model: str,
        embedding_model_size: int,
        enable_caching: bool = True,
        inline_tool_calls: bool = False,
        # Context Management
        max_assistant_loops: int,
        context_refresh_trigger_tokens: int,
        context_refresh_target_tokens: int,
        max_context_age_minutes: float,
        min_convo_age_for_greeting_minutes: float,
        enable_assistant_greeting: bool,
        initial_context_refresh_wait_seconds: int,
        # Memory Management
        memory_cluster_similarity_threshold: float,
        max_memory_cluster_size: int,
        min_memory_cluster_size: int,
        memories_between_consolidation: int,
        l2_memory_relevance_distance_threshold: float,
        # Basic Configuration
        debug: bool,
        default_persona: str,  # The generic persona to use if no persona is specified
        default_assistant_name: str,  # The generic assistant name to use if no assistant name is specified
    ):

        self.params = SimpleNamespace(**{k: v for k, v in locals().items() if k != "self"})

        self.user_token = user_token
        self.show_internal_thought = show_internal_thought
        self.default_assistant_name = default_assistant_name
        self.default_persona = default_persona
        self.enable_assistant_greeting = enable_assistant_greeting
        self.debug = debug
        self.context_refresh_trigger_tokens = context_refresh_trigger_tokens
        self.max_assistant_loops = max_assistant_loops
        self.context_refresh_trigger_tokens = context_refresh_trigger_tokens
        self.l2_memory_relevance_distance_threshold = l2_memory_relevance_distance_threshold
        self.context_refresh_target_tokens = context_refresh_target_tokens
        self.memory_cluster_similarity_threshold = memory_cluster_similarity_threshold
        self.min_memory_cluster_size = min_memory_cluster_size
        self.max_memory_cluster_size = max_memory_cluster_size
        self.memories_between_consolidation = memories_between_consolidation

    from ..tools.function_caller import ToolRegistry

    @cached_property
    def tool_registry(self) -> ToolRegistry:
        from ..tools.function_caller import ToolRegistry

        registry = ToolRegistry(self.params.custom_tools_path)
        registry.register_all()
        return registry

    @cached_property
    def config_path(self) -> Path:
        if self.params.config_path:
            return Path(self.params.config_path)
        else:
            return get_default_config_path()

    @property
    def initial_refresh_wait(self) -> timedelta:
        return timedelta(seconds=self.params.initial_context_refresh_wait_seconds)

    @property
    def max_in_context_message_age(self) -> timedelta:
        return timedelta(minutes=self.params.max_context_age_minutes)

    @property
    def min_convo_age_for_greeting(self) -> timedelta:
        return timedelta(minutes=self.params.min_convo_age_for_greeting_minutes)

    @cached_property
    def chat_model(self) -> ChatModel:
        return get_chat_model(
            model_name=self.params.chat_model,
            openai_api_key=self.params.openai_api_key,
            anthropic_api_key=self.params.anthropic_api_key,
            api_base=self.params.openai_api_base,
            organization=self.params.openai_organization,
            enable_caching=self.params.enable_caching,
            inline_tool_calls=self.params.inline_tool_calls,
        )

    @cached_property
    def embedding_model(self) -> EmbeddingModel:
        return get_embedding_model(
            model_name=self.params.embedding_model,
            embedding_size=self.params.embedding_model_size,
            api_key=self.params.openai_api_key,
            api_base=self.params.openai_api_base,
            organization=self.params.openai_organization,
            enable_caching=self.params.enable_caching,
        )

    @cached_property
    def is_new_user(self) -> bool:
        # This will be set when user_id property is accessed
        if not hasattr(self, "_is_new_user"):
            raise ValueError("Cannot determine if new user created, fetch user id first")
        return self._is_new_user

    @cached_property
    def user_id(self) -> int:
        stored_user_id = get_user_id_if_exists(self.db, self.user_token)

        if stored_user_id:
            self._is_new_user = False
            self._user_id = stored_user_id
        else:
            self._user_id = create_user_id(self.db, self.user_token)
            self._is_new_user = True
        return self._user_id

    @property
    def io(self) -> ElroyIO:
        if not self._io:
            self._io = CliIO(
                show_internal_thought=self.params.show_internal_thought,
                system_message_color=self.params.system_message_color,
                assistant_message_color=self.params.assistant_color,
                user_input_color=self.params.user_input_color,
                warning_color=self.params.warning_color,
                internal_thought_color=self.params.internal_thought_color,
            )
        return self._io

    def set_io(self, io: ElroyIO) -> None:
        self._io = io

    @property
    def db(self) -> DbManager:
        if not self._db:
            raise ValueError("No db session open")
        else:
            return self._db

    def is_db_connected(self) -> bool:
        return bool(self._db)

    @contextmanager
    def dbsession(self) -> Generator[None, None, None]:
        """Context manager for database sessions"""
        assert self.params.database_url, "Database URL not set"

        if self.params.database_url.startswith("postgresql://"):
            db_manager = PostgresManager
        elif self.params.database_url.startswith("sqlite:///"):
            db_manager = SqliteManager
        else:
            raise ValueError(f"Unsupported database URL: {self.params.database_url}. Must be either a postgresql:// or sqlite:/// URL")

        with db_manager.open_session(self.params.database_url, True) as db:
            try:
                self._db = db
                yield
            finally:
                self._db = None


T = TypeVar("T", bound=Callable[..., Any])


def get_ctx(typer_ctx: typer.Context) -> ElroyContext:
    ctx = typer_ctx.obj
    assert isinstance(ctx, ElroyContext), f"Context must be ElroyContext, got {type(ctx)}"
    return ctx


def clone_ctx_with_db(ctx: ElroyContext, db: DbManager) -> ElroyContext:
    new_ctx = ElroyContext(**vars(ctx.params))  # type: ignore
    new_ctx._db = db
    return new_ctx



import logging
from typing import Iterator, Union

from rich.console import Console, RenderableType

from ..db.db_models import FunctionCall
from ..llm.stream_parser import (
    AssistantInternalThought,
    SystemMessage,
    SystemWarning,
    TextOutput,
)


class ElroyIO:
    console: Console

    def print_stream(self, messages: Iterator[Union[TextOutput, RenderableType, FunctionCall]]) -> None:
        for message in messages:
            self.print(message, end="")
        self.console.print("")

    def print(self, message: Union[TextOutput, RenderableType, str, FunctionCall], end: str = "\n") -> None:
        if hasattr(message, "__rich__") or isinstance(message, str):
            self.console.print(message, end)
        else:
            raise NotImplementedError(f"Invalid message type: {type(message)}")

    def internal_thought(self, message: str):
        self.print(AssistantInternalThought(message))

    def info(self, message: Union[str, RenderableType]):
        if isinstance(message, str):
            self.print(SystemMessage(message))
        else:
            self.print(message)

    def warning(self, message: Union[str, RenderableType]):
        if isinstance(message, str):
            self.print(SystemWarning(message))
        else:
            self.print(message)


class StdIO(ElroyIO):
    """
    IO which emits plain text to stdin and stdout.
    """

    def __init__(self):
        self.console = Console(no_color=True)

    def print(self, message: Union[TextOutput, RenderableType, str, FunctionCall], end: str = "\n") -> None:
        if isinstance(message, AssistantInternalThought):
            logging.info(f"{type(message)}: {message}")
        elif isinstance(message, SystemWarning):
            logging.warning(message)
        elif isinstance(message, FunctionCall):
            logging.info(f"FUNCTION CALL: {message.function_name}({message.arguments})")
        elif isinstance(message, TextOutput):
            self.console.print(message.content, end=end)
        else:
            super().print(message, end)



import logging
import os
from functools import lru_cache
from pathlib import Path
from typing import Any, Callable, Dict, Optional

import yaml
from toolz import assoc, merge, pipe
from toolz.curried import map, valfilter
from typer import Option

from ..config.config import DEFAULTS_CONFIG
from ..config.models import resolve_anthropic
from ..config.paths import get_default_sqlite_url


def resolve_model_alias(alias: str) -> Optional[str]:
    if alias in ["sonnet", "opus", "haiku"]:
        return resolve_anthropic(alias)
    else:
        return {
            "gpt4o": "gpt-4o",
            "gpt4o_mini": "gpt-4o-mini",
            "o1": "o1",
            "o1_mini": "o1-mini",
        }.get(alias)


@lru_cache
def load_config_file_params(config_path: Optional[str] = None) -> Dict:
    # Looks for user specified config path, then merges with default values packaged with the lib

    user_config_path = config_path or os.environ.get(get_env_var_name("config_path"))

    if not user_config_path:
        return {}
    else:

        if user_config_path and not Path(user_config_path).is_absolute():
            logging.info("Resolving relative user config path")
            # convert to absolute path if not already, relative to working dir
            user_config_path = Path(user_config_path).resolve()
        return load_config_if_exists(user_config_path)


def ElroyOption(key: str, rich_help_panel: str, help: str, default_factory: Optional[Callable] = None, *args):
    """
    Typer options that have values in the user config file

    Creates a typer Option with value priority:
    1. CLI provided value (handled by typer)
    2. User config file value (if provided)
    3. defaults.yml value
    """

    return Option(
        *args,
        default_factory=default_factory if default_factory else lambda: load_config_file_params().get(key),
        envvar=get_env_var_name(key),
        rich_help_panel=rich_help_panel,
        help=help,
        show_default=str(DEFAULTS_CONFIG.get(key)),
    )


def get_env_var_name(parameter_name: str):
    return {
        "openai_api_key": "OPENAI_API_KEY",
        "openai_api_base": "OPENAI_API_BASE",
        "openai_embedding_api_base": "OPENAI_API_BASE",
        "openai_organization": "OPENAI_ORGANIZATION",
        "anthropic_api_key": "ANTHROPIC_API_KEY",
    }.get(parameter_name, "ELROY_" + parameter_name.upper())


def get_resolved_params(**kwargs) -> Dict[str, Any]:
    """Get resolved parameter values from environment and config."""
    # n.b merge priority is lib default < user config file < env var < explicit CLI arg

    return pipe(
        [
            DEFAULTS_CONFIG,  # package defaults
            load_config_file_params(kwargs.get("config_path")),  # user specified config file
            {k: os.environ.get(get_env_var_name(k)) for k in DEFAULTS_CONFIG.keys()},  # env vars
            kwargs,  # explicit params
        ],
        map(valfilter(lambda x: x is not None)),
        merge,
        lambda d: assoc(d, "database_url", get_default_sqlite_url()) if not d.get("database_url") else d,
    )  # type: ignore


@lru_cache
def load_config_if_exists(user_config_path: Optional[str]) -> dict:
    """
    Load configuration values in order of precedence:
    1. defaults.yml (base defaults)
    2. User config file (if provided)
    """

    if not user_config_path:
        return {}

    if not Path(user_config_path).exists():
        logging.info(f"User config file {user_config_path} not found")
        return {}
    elif not Path(user_config_path).is_file():
        logging.error(f"User config path {user_config_path} is not a file")
        return {}
    else:
        try:
            with open(user_config_path, "r") as user_config_file:
                return yaml.safe_load(user_config_file)
        except Exception as e:
            logging.error(f"Failed to load user config file {user_config_path}: {e}")
            return {}



import json
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generator, Generic, Iterator, Optional, Type, TypeVar, Union

from litellm.types.utils import Delta, ModelResponse

from ..config.config import ChatModel
from ..db.db_models import FunctionCall
from .tool_call_accumulator import OpenAIToolCallAccumulator


@dataclass
class TextOutput(ABC):
    content: str


class AssistantInternalThought(TextOutput):
    content: str


class AssistantResponse(TextOutput):
    content: str


class AssistantToolResult(TextOutput):
    content: str


class SystemMessage(TextOutput):
    content: str


class SystemWarning(TextOutput):
    content: str


class InlineToolCall(TextOutput):
    content: str


def to_openai_tool_call(content: str) -> Optional[FunctionCall]:
    try:
        d = json.loads(content)
        if d.get("name") and d.get("arguments"):
            return FunctionCall(id=uuid.uuid4().hex, function_name=d["name"], arguments=d["arguments"])
    except Exception:
        pass


T = TypeVar("T", bound=TextOutput)


class TextAccumulator(Generic[T]):
    def __init__(self, output_type: Type[T], opening_tag: str, closing_tag: str):
        self.output_type = output_type
        self.opening_tag = opening_tag
        self.closing_tag = closing_tag

        self.is_active = False
        self.is_first_output_chunk_emitted = None
        self.buffer = ""

    def update(self, content_chunk: str) -> Generator[Union[T, str], None, None]:
        # Accepts text as input, one or more characters at a time.
        # If not active:
        #   # if the accumulated text might contain the beginning of the opening tag, accumulate text
        #   # if the accumulated text no longer can match the opening tag, yield the accumulated text as a string
        #   # if the accumulated text matches the opening tag, set is_active to True, and yield any text that follows the opening tag as AssistantInternalThought
        # If active:
        #   # if the accumulated text might contain the beginning of the closing tag, accumulate text
        #   # if the accumulated text no longer can match the closing tag, yield the accumulated text as a string
        #   # if the accumulated text matches the closing tag, set is_active to False, and yield any text that precedes the closing tag as AssistantInternalThought, yield any text that follows the closing tag as a string
        self.buffer += content_chunk

        while self.buffer:
            if self.buffer.isspace():
                break
            elif self.is_active:
                if self.closing_tag in self.buffer:
                    text_before_tag, text_after_tag = self.buffer.split(self.closing_tag, maxsplit=1)

                    if text_before_tag and not text_before_tag.isspace():  # if we have non-space text before the closing tag
                        if not self.is_first_output_chunk_emitted:
                            content = text_before_tag.strip()
                            self.is_first_output_chunk_emitted = True
                        else:
                            content = text_before_tag.rstrip()
                        yield self.output_type(content)
                    self.is_active = False
                    self.buffer = text_after_tag
                    self.is_first_output_chunk_emitted = None
                elif self._has_possible_tag_fragment(self.buffer, self.closing_tag):
                    break
                else:
                    if not self.is_first_output_chunk_emitted:
                        content = self.buffer.lstrip()
                        self.is_first_output_chunk_emitted = True
                    else:
                        content = self.buffer
                    yield self.output_type(content)
                    self.buffer = ""
            else:
                if self.opening_tag in self.buffer:
                    self.is_active = True
                    self.is_first_output_chunk_emitted = False
                    text_before_tag, text_after_tag = self.buffer.split(self.opening_tag, maxsplit=1)

                    if text_before_tag:
                        yield text_before_tag.rstrip()
                    self.buffer = text_after_tag

                elif self._has_possible_tag_fragment(self.buffer, self.opening_tag):
                    break
                else:
                    yield self.buffer
                    self.buffer = ""

    def _has_possible_tag_fragment(self, text: str, tag: str) -> bool:
        if tag in text:
            raise ValueError("This function only intended to capture partial matches, a full match indicates a bug")

        while tag[0] in text:
            text = text[text.index(tag[0]) :]

            if tag.startswith(text[: len(tag)]):
                return True
            else:
                text = text[1:]
        return False

    def flush(self) -> Generator[Union[T, str], None, None]:
        if self.buffer:
            yield self.buffer
            self.buffer = ""
            self.is_active = False


class TextProcessor(ABC):
    def __init__(self, next_processor: Optional["TextProcessor"] = None):
        self.next_processor = next_processor

    @abstractmethod
    def process(self, text: str) -> Generator[Union[TextOutput, FunctionCall], None, None]:
        pass

    def process_next(self, text: str) -> Generator[Union[TextOutput, FunctionCall], None, None]:
        if self.next_processor:
            yield from self.next_processor.process(text)
        else:
            yield AssistantResponse(text)

    def flush(self) -> Generator[Union[TextOutput, FunctionCall], None, None]:
        if self.next_processor:
            yield from self.next_processor.flush()


class InternalThoughtProcessor(TextProcessor):
    def __init__(self, next_processor: Optional[TextProcessor] = None):
        super().__init__(next_processor)
        self.accumulator = TextAccumulator(AssistantInternalThought, "<internal_thought>", "</internal_thought>")

    def process(self, text: str) -> Generator[Union[TextOutput, FunctionCall], None, None]:
        for processed_text in self.accumulator.update(text):
            if isinstance(processed_text, AssistantInternalThought):
                yield processed_text
            else:
                yield from self.process_next(processed_text)

    def flush(self) -> Generator[Union[TextOutput, FunctionCall], None, None]:
        for processed_text in self.accumulator.flush():
            if isinstance(processed_text, AssistantInternalThought):
                yield processed_text
            else:
                yield from self.process_next(processed_text)
        if self.next_processor:
            yield from self.next_processor.flush()


# TODO: This needs update needs to be handled differently, we must accumulate all the text from the accumulator until we have a function call, then yield.
class InlineToolCallProcessor(TextProcessor):
    def __init__(self, next_processor: Optional[TextProcessor] = None):
        super().__init__(next_processor)
        self.accumulator = TextAccumulator(InlineToolCall, "<tool_call>", "</tool_call>")
        self.tool_call_text = ""

    def process(self, text: str) -> Generator[Union[TextOutput, FunctionCall], None, None]:
        for processed_text in self.accumulator.update(text):
            if isinstance(processed_text, InlineToolCall):
                self.tool_call_text += processed_text.content
                tool_call = to_openai_tool_call(self.tool_call_text)
                if tool_call:
                    yield tool_call
                    self.tool_call_text = ""
            else:
                yield from self.process_next(processed_text)


class StreamParser:
    def __init__(self, chat_model: ChatModel, chunks: Iterator[ModelResponse]):
        self.chunks = chunks
        self.openai_tool_call_accumulator = OpenAIToolCallAccumulator(chat_model)
        # Chain the processors
        self.text_processor = InternalThoughtProcessor(InlineToolCallProcessor())
        self.raw_text = None

    def process(self) -> Generator[Union[TextOutput, FunctionCall], None, None]:
        for chunk in self.chunks:
            delta = chunk.choices[0].delta  # type: ignore
            assert isinstance(delta, Delta)
            if delta.tool_calls:
                yield from self.openai_tool_call_accumulator.update(delta.tool_calls)
            if delta.content:
                text = delta.content
                if not self.raw_text:
                    self.raw_text = text
                else:
                    self.raw_text += text
                assert isinstance(text, str)
                yield from self.process_text_chunk(text)
        yield from self.text_processor.flush()

    def get_full_text(self):
        return self.raw_text

    def process_text_chunk(self, text: str) -> Generator[Union[TextOutput, FunctionCall], None, None]:
        yield from self.text_processor.process(text)



from datetime import datetime
from functools import wraps
from typing import Callable, Generator, Optional

from pytz import UTC

from .cli.options import get_resolved_params
from .config.constants import USER
from .config.ctx import ElroyContext
from .llm.persona import get_persona as do_get_persona
from .llm.stream_parser import AssistantInternalThought
from .messaging.messenger import process_message
from .repository.memories.operations import create_memory
from .tools.user_preferences import set_assistant_name, set_persona


def db(f: Callable) -> Callable:
    """Decorator to wrap function calls with database session context"""

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        if not self.ctx.is_db_connected():
            with self.ctx.dbsession():
                return f(self, *args, **kwargs)

    return wrapper


class Elroy:
    ctx: ElroyContext

    def __init__(
        self,
        token: Optional[str] = None,
        config_path: Optional[str] = None,
        persona: Optional[str] = None,
        assistant_name: Optional[str] = None,
        database_url: Optional[str] = None,
    ):

        self.ctx = ElroyContext(
            **get_resolved_params(
                user_token=token,
                config_path=config_path,
                database_url=database_url,
            ),
        )
        with self.ctx.dbsession():
            if persona:
                set_persona(self.ctx, persona)

            if assistant_name:
                set_assistant_name(self.ctx, assistant_name)

    @db
    def message(self, input: str) -> str:
        return "".join(self.message_stream(input))

    def message_stream(self, input: str) -> Generator[str, None, None]:
        stream = [
            chunk.content
            for chunk in process_message(USER, self.ctx, input)
            if not isinstance(chunk, AssistantInternalThought) or self.ctx.show_internal_thought
        ]
        if not self.ctx.is_db_connected():
            with self.ctx.dbsession():
                yield from stream
        else:
            yield from stream

    @db
    def remember(self, message: str, name: Optional[str] = None) -> None:
        if not name:
            name = f"Memory from {datetime.now(UTC)}"
        create_memory(self.ctx, name, message)

    @db
    def get_persona(self) -> str:
        return do_get_persona(self.ctx)



import os
from pathlib import Path

APP_NAME = "elroy"
APP_AUTHOR = "elroy-bot"


def get_home_dir() -> Path:
    """Get the Elroy home directory (~/.elroy), creating it if it doesn't exist.

    Can be overridden with ELROY_HOME environment variable.
    """
    if env_home := os.environ.get("ELROY_HOME"):
        home_dir = Path(env_home)
    else:
        home_dir = Path.home() / ".elroy"

    home_dir.mkdir(parents=True, exist_ok=True)
    return home_dir


def get_default_config_path() -> Path:
    return get_home_dir() / "elroy.conf.yaml"


def get_cache_dir() -> Path:
    cache_dir = get_home_dir() / "cache"
    cache_dir.mkdir(parents=True, exist_ok=True)
    return cache_dir


def get_prompt_history_path():
    return get_cache_dir() / "history"


def get_default_sqlite_url():
    return f"sqlite:///{get_home_dir() / 'elroy.db'}"


def get_log_file_path():
    logs_dir = get_home_dir() / "logs"
    logs_dir.mkdir(parents=True, exist_ok=True)
    return logs_dir / "elroy.log"
